#! /usr/bin/env python3
"""test_histFit.py
   histogram fit  with iminiut

.. moduleauthor:: Guenter Quast <g.quast@kit.edu>

"""

import numpy as np, matplotlib.pyplot as plt
from PhyPraKit.phyFit import hFit
from kafe2 import Fit, Plot, HistContainer


if __name__ == "__main__": # --------------------------------------  
  #
  # Example of a histogram fit
  #

  # define the model function to fit
  def model(x, mu = 6.0, sigma = 0.5, s = 0.3):
    '''pdf of a Gaussian signal on top of flat background
    '''
    normal = np.exp(-0.5*((x-mu)/sigma)**2)/np.sqrt(2.*np.pi*sigma**2)
    flat = 1./(max-min) 
    return s * normal + (1-s) * flat

  #
  # generate Histogram Data
  #
  # parameters of data sample, signal and background parameters
  N = 100  # number of entries
  min=0.   # range of data, mimimum
  max= 10. # maximum
  s = 0.25 # signal fraction 
  pos = 6.66 # signal position
  width = 0.33 # signal width
  
  # fix random generator seed 
  np.random.seed(314159)  # initialize random generator

  def generate_data(N, min, max, p, w, s):
    '''generate a random dataset: 
       Gaussian signal at position p with width w and signal fraction s 
       on top of a flat background between min and max
     '''
     # signal sample
    data_s = np.random.normal(loc=pos, scale=width, size=int(s*N) )
     # background sample
    data_b = np.random.uniform(low=min, high=max, size=int((1-s)*N) )
    return np.concatenate( (data_s, data_b) )

  # generate a data sample ...
  SplusB_data = generate_data(N, min, max, pos, width, s)

  # ... define binning for analysis ...
  nbins=40
  bedges=np.linspace(min, max, nbins+1)
# ... and create a histogram
  bc, be = np.histogram(SplusB_data, bins=bedges)
  print (bc)
  
#  
# ---  perform fit  
#
  parvals, parerrs, cor, gof = hFit(model,
      bc, be,              # bin entries and bin edges
      p0=None,        # initial guess for parameter values 
   #  constraints=['name', val ,err ],   # constraints within errors
      limits=('s', 0., None),  #limits
      use_GaussApprox=False,   # Gaussian approximation
      fit_density = True,      # fit density
      plot=True,           # plot data and model
      plot_band=True,      # plot model confidence-band
      plot_cor=False,      # plot profiles likelihood and contours
      showplots=False,      # show / don't show plots
      quiet=False,         # suppress informative printout
      axis_labels=['x', 'y   \  f(x, *par)'], 
      data_legend = 'random data',    
      model_legend = 'signal + background model'
  )

# Print results to illustrate how to use output
  print('\n*==* Fit Result:')
  print(" goodness-of-fit: {:.3g}".format(gof))
  print(" parameter values:      ", parvals)
  print(" neg. parameter errors: ", parerrs[:,0])
  print(" pos. parameter errors: ", parerrs[:,1])
  print(" correlations : \n", cor)
  
#
# compare with kafe2 histogram fit
#
  k2HContainer =  HistContainer(n_bins=nbins,
                                   bin_range=(min, max), fill_data=SplusB_data)
  k2h_fit = Fit(data=k2HContainer, model_function=model)
  k2h_fit.do_fit()  # do the fit
  k2h_plot = Plot(k2h_fit)
  k2h_plot.plot(asymmetric_parameter_errors=True)

  plt.show()

  
